@@ -1,4 +1,7 @@
 
-__version__ = "0.1.7"
+__version__ = "0.1.11"
 docs = r"https://alexlib.readthedocs.io/en/latest/"
 git = r"https://github.com/thisismygitrepo/alexlib"
+release_notes = """
+Created Experimental Class
+"""
