@@ -1,5 +1,3 @@
-
-
 import alexlib.toolbox as tb
 import numpy as np
 from abc import ABC, abstractmethod
@@ -26,26 +24,33 @@ class HyperParam(tb.Struct):
     * When doing multiple experiments, one command in console reminds you of settings used in that run (hp.__dict__).
     * Ease of saving settings of experiments! and also replicating it later.
     """
-    subpath = 'metadata/HyperParam.pickle'
+    subpath = 'metadata/HyperParam.pkl'
 
-    def __init__(self, **kwargs):
-        super(HyperParam, self).__init__(**kwargs)
-        """
-        """
-        # ==================== Enviroment ========================
-        self.exp_name = 'default'
-        self.root = 'tmp'
-        self.pkg_name = None
-        self.device_name = Device.gpu0
-        # ===================== DATA ============================
-        self.seed = 234
-        # ===================== Model =============================
-        # ===================== Training ========================
-        self.split = 0.2
-        self.lr = 0.0005
-        self.batch_size = 32
-        self.epochs = 30
+    def __init__(self, *args, **kwargs):
+        super().__init__(
+            # ==================== Enviroment =========================
+            exp_name='default',
+            root='tmp',
+            pkg_name='tensorflow',
+            device_name=Device.gpu0,
+            # ===================== Data ==============================
+            seed=234,
+            shuffle=True,
+            precision='float32',
+            # ===================== Model =============================
+            # ===================== Training ==========================
+            split=0.2,
+            lr=0.0005,
+            batch_size=32,
+            epochs=30,
+        )
         self._code = None
+        self._configured = False
+        self.root = None
+        self.device_name = None
+        self.update(*args, **kwargs)
+        # self.save_code()
+        # self.config_device()
 
     def save_code(self):
         import inspect
@@ -80,7 +85,50 @@ class HyperParam(tb.Struct):
 
     @property
     def device(self):
-        return self.config_device(self.pkg, self.device_name)
+        handle = self.pkg
+        if handle.__name__ == 'tensorflow':
+            """
+            To disable gpu, here's one way: # before importing tensorflow do this:
+            if device == 'cpu':
+                os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
+            handle.device(device)  # used as context, every tensor constructed and every computation takes place therein
+            For more manual control, use .cpu() and .gpu('0') .gpu('1') attributes.
+            """
+            devices = handle.config.experimental.list_physical_devices('CPU')
+            devices += handle.config.experimental.list_physical_devices('GPU')
+            device_dict = dict(zip(['cpu', 'gpu0', 'gpu1'], devices))
+
+            if self.device_name is Device.auto:
+                chosen_device = Device.gpu0 if len(devices) > 1 else Device.cpu
+            else:
+                chosen_device = self.device_name
+
+            device_str = chosen_device.value if 1 > 0 else "haha"
+            assert device_str in device_dict.keys(), f"This machine has no such a device to be chosen! ({device_str})"
+
+            try:
+                device = device_dict[device_str]
+                return device
+            except KeyError:  # 2gpus not a key in the dict.
+                assert len(handle.config.experimental.get_visible_devices()) > 2
+                mirrored_strategy = handle.distribute.MirroredStrategy()
+                return mirrored_strategy
+
+        elif handle.__name__ == 'torch':
+            device = self.device_name
+            if device is Device.auto:
+                return handle.device('cuda:0') if handle.cuda.is_available() else handle.device('cpu')
+            elif device is Device.gpu0:
+                assert handle.cuda.device_count() > 0, f"GPU {device} not available"
+                return handle.device('cuda:0')
+            elif device is Device.gpu1:
+                assert handle.cuda.device_count() > 1, f"GPU {device} not available"
+                return handle.device('cuda:1')
+            elif device is Device.cpu:
+                return handle.device('cpu')
+            # How to run Torch model on 2 GPUs ?
+        else:
+            raise NotImplementedError(f"I don't know how to configure devices for this package {handle}")
 
     @property
     def pkg(self):
@@ -91,75 +139,44 @@ class HyperParam(tb.Struct):
             handle = __import__("torch")
         return handle
 
-    @staticmethod
-    def config_device(handle, device: Device = Device.gpu0):
+    def config_device(self):
         """
-        :param handle: package handle
-        :param device: device
-        :return: possibly a handle to device (in case of Pytorch)
         """
+        handle = self.pkg
+        device_str = self.device_name.value
+        device = self.device
+        if handle.__name__ == 'torch':
+            return None
+
         try:
-            if handle.__name__ == 'tensorflow':
-                """
-                To disable gpu, here's one way: # before importing tensorflow do this:
-                if device == 'cpu':
-                    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
-                handle.device(device)  # used as context, every tensor constructed and every computation takes place therein
-                For more manual control, use .cpu() and .gpu('0') .gpu('1') attributes.
-                """
-                devices = handle.config.experimental.list_physical_devices('CPU')
-                devices += handle.config.experimental.list_physical_devices('GPU')
-                device_dict = dict(zip(['cpu', 'gpu0', 'gpu1'], devices))
+            # Now we want only one device to be seen:
+            if device_str in ['gpu0', 'gpu1']:
+                limit_memory = True
+                if limit_memory:  # memory growth can only be limited for GPU devices.
+                    handle.config.experimental.set_memory_growth(device, True)
+                handle.config.experimental.set_visible_devices(device, 'GPU')  # will only see this device
+                # logical_gpus = handle.config.experimental.list_logical_devices('GPU')
+                # now, logical gpu is created only for visible device
+                # print(len(devices), "Physical devices,", len(logical_gpus), "Logical GPU")
+            else:  # for cpu devices, we want no gpu to be seen:
+                handle.config.experimental.set_visible_devices([], 'GPU')  # will only see this device
+                # logical_gpus = handle.config.experimental.list_logical_devices('GPU')
+                # now, logical gpu is created only for visible device
+                # print(len(devices), "Physical devices,", len(logical_gpus), "Logical GPU")
 
-                if device is Device.auto:
-                    device = Device.gpu0 if len(devices) > 1 else Device.cpu
-
-                device_str = device.value if 1 > 0 else "haha"
-                assert device_str in device_dict.keys(), f"This machine has no such a device to be chosen! ({device_str})"
-                if device_str != '2gpus':
-                    device = device_dict[device_str]
-                    # Now we want only one device to be seen:
-                    if device_str in ['gpu0', 'gpu1']:
-                        limit_memory = True
-                        if limit_memory:  # memory growth can only be limited for GPU devices.
-                            handle.config.experimental.set_memory_growth(device, True)
-                        handle.config.experimental.set_visible_devices(device, 'GPU')  # will only see this device
-                        logical_gpus = handle.config.experimental.list_logical_devices('GPU')
-                        # now, logical gpu is created only for visible device
-                        print(len(devices), "Physical devices,", len(logical_gpus), "Logical GPU")
-                    else:  # for cpu devices, we want no gpu to be seen:
-                        handle.config.experimental.set_visible_devices([], 'GPU')  # will only see this device
-                        logical_gpus = handle.config.experimental.list_logical_devices('GPU')
-                        # now, logical gpu is created only for visible device
-                        print(len(devices), "Physical devices,", len(logical_gpus), "Logical GPU")
-                    return device
-                else:
-                    assert len(handle.config.experimental.get_visible_devices()) > 2
-                    mirrored_strategy = handle.distribute.MirroredStrategy()
-                    return mirrored_strategy
-
-            elif handle.__name__ == 'torch':
-                if device is Device.auto:
-                    return handle.device('cuda:0') if handle.cuda.is_available() else handle.device('cpu')
-                elif device is Device.gpu0:
-                    assert handle.cuda.device_count() > 0, f"GPU {device} not available"
-                    return handle.device('cuda:0')
-                elif device is Device.gpu1:
-                    assert handle.cuda.device_count() > 1, f"GPU {device} not available"
-                    return handle.device('cuda:1')
-                elif device is Device.cpu:
-                    return handle.device('cpu')
-                # How to run Torch model on 2 GPUs ?
-            else:
-                raise NotImplementedError(f"I don't know how to configure devices for this package {handle}")
         except AssertionError as e:
             print(e)
             print(f"Trying again with auto-device {Device.auto}")
-            HyperParam.config_device(handle, device=Device.auto)
+            self.device_name = Device.auto
+            self.config_device()
+
+        except RuntimeError as e:
+            print(e)
+            print(f"Device already configured, skipping ... ")
 
 
 class DataReader(tb.Base):
-    subpath = "metadata/DataReader.pickle"
+    subpath = "metadata/DataReader.pkl"
     """This class holds the dataset for training and testing. However, it also holds meta data for preprocessing
     and postprocessing. The latter is essential at inference time, but the former need not to be saved. As such,
     at save time, this class only remember the attributes inside `.data_specs` `Struct`. Thus, whenever encountering
@@ -167,11 +184,13 @@ class DataReader(tb.Base):
     implemented a fallback `getattr` method that allows accessing those attributes from the class itself, without the 
     need to reference `.dataspects`.
     """
+
     def __init__(self, hp=None, data_specs=None, split=None, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.hp = hp
         self.data_specs = data_specs if data_specs else tb.Struct()  # Summary of data to be memorized by model
         self.split = split
+        self.plotter = None
 
     def __getattr__(self, item):
         try:
@@ -195,17 +214,19 @@ class DataReader(tb.Base):
         :param strings:
         :return:
         """
-        import sklearn.preprocessing as preprocessing
+        # import sklearn.preprocessing as preprocessing
         from sklearn.model_selection import train_test_split
         result = train_test_split(*args, test_size=self.hp.split, shuffle=self.hp.shuffle,
                                   random_state=self.hp.seed, **kwargs)
         self.split = tb.Struct(train_loader=None, test_loader=None)
+        if strings is None:
+            strings = ["x", "y"]
         self.split.update({astring + '_train': result[ii * 2] for ii, astring in enumerate(strings)})
         self.split.update({astring + '_test': result[ii * 2 + 1] for ii, astring in enumerate(strings)})
 
     def get_data_tuple(self, aslice, dataset="test"):
         # returns a tuple containing a slice of data (x_test, x_test, names_test, index_test etc)
-        keys = self.split.keys().filter(f"'_{train}' in x")
+        keys = self.split.keys().filter(f"'_{dataset}' in x")
         return tuple([self.split[key][aslice] for key in keys])
 
     def save_pickle(self, path=None, *names, **kwargs):
@@ -241,15 +262,28 @@ class DataReader(tb.Base):
 
     def preprocess(self, *args, **kwargs):
         _ = args, kwargs, self
-        return args
+        return args[0]  # acts like identity.
 
     def postprocess(self, *args, **kwargs):
         _ = args, kwargs, self
-        return args
+        return args[0]  # acts like identity
+
+    def image_viz(self, pred, gt=None, names=None, **kwargs):
+        """
+        Assumes numpy inputs
+        """
+        if gt is None:
+            labels = None
+        else:
+            labels = ['Reconstruction', 'Ground Truth']
+        self.plotter = tb.ImShow(pred, gt, labels=labels, sup_titles=names, origin='lower', **kwargs)
+
+    def viz(self, *args, **kwargs):
+        return None
 
 
 class BaseModel(ABC):
-    f"""My basic model. It implements the following methods:
+    """My basic model. It implements the following methods:
 
     * :func:`BaseModel.preprocess` This should convert to tensors as appropriate for the model.
     * :func:`BaseModel.postprocess` This method should convert back to numpy arrays.
@@ -257,6 +291,7 @@ class BaseModel(ABC):
     * :func:`BaseModel.predict` expects a processed input, uese infer and does postprocessing.
     * :func:`BaseModel.predict_from_s` reads, preprocess, then uses predict method.
     * :func:`BseModel.evaluate` Expects processed input and internally calls infer and postprocess methods.
+    
     """
 
     @abstractmethod
@@ -273,8 +308,15 @@ class BaseModel(ABC):
     def compile(self, loss=None, optimizer=None, metrics=None, compile_model=True):
         """ Updates compiler attributes. This acts like a setter.
 
+        .. note:: * this method is as good as setting attributes of `compiler` directly in case of PyTorch.
+                  * In case of TF, this is not the case as TF requires actual futher different
+                    compilation before changes take effect.
+
+        Remember:
+
         * Must be run prior to fit method.
         * Can be run only after defining model attribute.
+
         """
         pkg = self.hp.pkg
         if self.hp.pkg_name == 'tensorflow':
@@ -301,12 +343,13 @@ class BaseModel(ABC):
 
     def fit(self, viz=False, **kwargs):
         default_settings = tb.Struct(x=self.data.split.x_train, y=self.data.split.y_train,
-                                validation_data=(self.data.split.x_test, self.data.split.y_test),
-                                batch_size=self.hp.batch_size, epochs=self.hp.epochs, verbose=1,
-                                shuffle=self.hp.shuffle, callbacks=[])
+                                     validation_data=(self.data.split.x_test, self.data.split.y_test),
+                                     batch_size=self.hp.batch_size, epochs=self.hp.epochs, verbose=1,
+                                     shuffle=self.hp.shuffle, callbacks=[])
         default_settings.update(kwargs)
-        hist = self.model.fit(**default_settings)
-        self.history.append(tb.Struct(tb.copy.deepcopy(hist.history)))  # it is paramount to copy, cause source can change.
+        hist = self.model.fit(**default_settings.dict)
+        self.history.append(tb.Struct(tb.copy.deepcopy(hist.history)))
+        # it is paramount to copy, cause source can change.
         if viz:
             self.plot_loss()
         return self
@@ -361,28 +404,17 @@ class BaseModel(ABC):
         inferred = self.infer(x)
         return self.postprocess(inferred, **kwargs)
 
-    def predict_from_position(self, position, viz=True, **kwargs):
-        s_obj = self.process_position(position)
-        self.predict_from_s_obj(s_obj, viz=viz, **kwargs)
-
-    def predict_from_s_obj(self, s_obj, viz=True, **kwargs):
-        preprocessed = self.preprocess(s_obj)
+    def deduce(self, obj_s, viz=True, **kwargs):
+        preprocessed = np.concatenate([self.preprocess(obj) for obj in obj_s], axis=0)
         prediction = self.infer(preprocessed)
         postprocessed = self.postprocess(prediction, **kwargs)
-        s_obj.prediction = tb.Struct(preprocessed=preprocessed, prediction=prediction, postprocessed=postprocessed)
+        result = tb.Struct(input=obj_s, preprocessed=preprocessed, prediction=prediction, postprocessed=postprocessed)
         if viz:
-            return self.viz(postprocessed, **kwargs)
-        return postprocessed
+            self.viz(postprocessed, **kwargs)
+        return result
 
-    def viz(self, pred, gt=None, names=None, **kwargs):
-        """
-        Assumes numpy inputs
-        """
-        if gt is None:
-            labels = None
-        else:
-            labels = ['Reconstruction', 'Ground Truth']
-        self.plotter = tb.ImShow(pred, gt, labels=labels, sup_titles=names, origin='lower', **kwargs)
+    def viz(self, *args, **kwargs):
+        self.data.viz(*args, **kwargs)
 
     def evaluate(self, x_test=None, y_test=None, names_test=None, idx=None, viz=True, sample=5, **kwargs):
         # ================= Data Procurement ===================================
@@ -392,9 +424,10 @@ class BaseModel(ABC):
         names_test = names_test if names_test is not None else this
         if idx is None:
             def get_rand(x, y):
-                idx_ = np.random.choice(len(x)-sample)
-                return x[idx_:idx_ + sample], y[idx_:idx_ + sample],\
-                       names_test[idx_: idx_ + sample], np.arange(idx_, idx_ + sample)
+                idx_ = np.random.choice(len(x) - sample)
+                return x[idx_:idx_ + sample], y[idx_:idx_ + sample], \
+                    names_test[idx_: idx_ + sample], np.arange(idx_, idx_ + sample)
+
             assert self.data is not None, 'Data attribute is not defined'
             x_test, y_test, names_test, idx = get_rand(x_test, y_test)  # already processed S's
         else:
@@ -408,36 +441,37 @@ class BaseModel(ABC):
 
         prediction = self.infer(x_test)
         loss_dict = self.get_metrics_evaluations(prediction, y_test)
+        loss_dict['names'] = names_test
         pred = self.postprocess(prediction, per_instance_kwargs=dict(name=names_test), legend="Prediction", **kwargs)
         gt = self.postprocess(y_test, per_instance_kwargs=dict(name=names_test), legend="Ground Truth", **kwargs)
         results = tb.Struct(pp_prediction=pred, prediction=prediction, input=x_test, pp_gt=gt, gt=y_test,
-                            names=names_test, losses=loss_dict,)
+                            names=names_test, loss_df=loss_dict, )
         if viz:
-            loss_name = results.losses.index(0)[0]
-            names = f" = {loss_name}\n".join(results.losses.index(0)[1].apply("x.numpy()").apply(str)).split("\n")
-            names = [f"{aname}. Case: {anindex}" for aname, anindex in zip(names, names_test)]
+            loss_name = results.loss_df.columns.to_list()[0]  # first loss name
+            loss_label = results.loss_df[loss_name].apply(lambda x: f"{loss_name} = {x}").to_list()
+            names = [f"{aname}. Case: {anindex}" for aname, anindex in zip(loss_label, names_test)]
             self.viz(pred, gt, names=names, **kwargs)
         return results
 
-    def get_metrics_evaluations(self, prediction, y_test):
-        if self.compiler is None or self.compiler.metrics is None:
+    def get_metrics_evaluations(self, prediction, groun_truth):
+        if self.compiler is None:
             return None
-        loss_dict = tb.Struct()
-        for a_metric in self.compiler.metrics:
-            stateful = False
+
+        metrics = tb.L([self.compiler.loss]) + self.compiler.metrics
+        loss_dict = dict()
+        for a_metric in metrics:
             try:  # EAFP principle.
                 name = a_metric.name  # works for subclasses Metrics
-                stateful = True
             except AttributeError:
                 name = a_metric.__name__  # works for functions.
-            loss_dict[name] = tb.List()
+            loss_dict[name] = []
 
-            for a_prediction, a_y_test in zip(prediction, y_test):
-                if stateful:
+            for a_prediction, a_y_test in zip(prediction, groun_truth):
+                if hasattr(a_metric, "reset_states"):
                     a_metric.reset_states()
                 loss = a_metric(a_prediction[None], a_y_test[None])
-                loss_dict[name].append(loss)
-        return loss_dict
+                loss_dict[name].append(np.array(loss).item())
+        return tb.pd.DataFrame(loss_dict)
 
     def save_model(self, directory):
         self.model.save(directory, include_optimizer=False)  # send only folder name. Save name is saved_model.pb
@@ -454,14 +488,14 @@ class BaseModel(ABC):
         name = directory.glob('*.data*').__next__().__str__().split('.data')[0]
         self.model.load_weights(name)  # requires path to file name.
 
-    def save_class(self, weights_only=True, version='0', itself=False, **kwargs):
+    def save_class(self, weights_only=True, save_sourcecode=True, version='0', itself=False, **kwargs):
         """Simply saves everything:
 
         1. Hparams
         2. Data specs
         3. Model architecture or weights depending on the following argument.
 
-        :param version:
+        :param version: Model version, up to the user.
         :param weights_only: self-explanatory
         :return:
 
@@ -481,9 +515,11 @@ class BaseModel(ABC):
         code_string = inspect.getsource(self.__class__)
         meta_dir = tb.P(self.hp.save_dir).joinpath('metadata').create()
         meta_dir.joinpath('model_arch.txt').write_text(code_string)
+        if save_sourcecode:
+            tb.P(inspect.getmodule(self).__file__).zip(op_path=meta_dir / "source_code")
 
         self.history.save_npy(path=meta_dir.joinpath('history.npy'))
-        print(f'Model calss saved successfully!, check out: \n {self.hp.save_dir}')
+        print(f'Model class saved successfully!, check out: \n {self.hp.save_dir.as_uri()}')
 
     @classmethod
     def from_class_weights(cls, path, hp_class=None, data_class=None, device_name=None):
@@ -492,7 +528,8 @@ class BaseModel(ABC):
         if hp_class:
             hp_obj = hp_class.from_saved(path)
         else:
-            print(f"HParam class not passed to constructor, assuming it is a self-sufficient save.")
+            print("<" * 1000, f"HParam class not passed to constructor, assuming it is a self-contained save.",
+                  "<" * 1000)
             hp_obj = tb.Read.pickle(path / HyperParam.subpath)
         if device_name:
             hp_obj.device_name = device_name
@@ -500,7 +537,8 @@ class BaseModel(ABC):
         if data_class:
             data_obj = data_class.from_saved(path, hp_obj) if path.exists() else None
         else:
-            print(f"Data class not passed to constructor, assuming it is a self-sufficient save.")
+            print("<" * 1000, f"Data class not passed to constructor, assuming it is a self-contained save.",
+                  "<" * 1000)
             data_path = path / DataReader.subpath
             data_obj = tb.Read.pickle(data_path) if data_path.exists() else None
         model_obj = cls(hp_obj, data_obj)
@@ -515,7 +553,7 @@ class BaseModel(ABC):
         path = tb.P(path)
         data_obj = DataReader.from_saved(path)
         hp_obj = HyperParam.from_saved(path)
-        model_obj = cls.load_model(path.myglob('/*_save_*')[0])  # static method.
+        model_obj = cls.load_model(path.myglob('*_save_*')[0])  # static method.
         tmp = cls.__init__
 
         def initializer(self, hp_object, data_object, model_object_):
@@ -542,14 +580,14 @@ class BaseModel(ABC):
                                   show_shapes=True, show_layer_names=True, **kwargs)
         print('Successfully plotted the model')
 
-    def build(self, shape=None):
+    def build(self, shape=None, verbose=True):
         """ Building has two main uses.
 
         * Useful to baptize the model, especially when its layers are built lazily. Although this will eventually
           happen as the first batch goes in. This is a must before showing the summary of the model.
         * Doing sanity check about shapes when designing model.
         * Sanity check about values and ranges when random normal input is fed.
-        :param dtype:
+
         :param shape:
         :return:
         """
@@ -562,12 +600,13 @@ class BaseModel(ABC):
         ip = np.random.randn(*((self.hp.batch_size,) + shape)).astype(dtype)
         op = self.model(ip)
         self.tmp = op
-        print("============  Build Test ==============")
-        print(f"Input shape = {ip.shape}")
-        print(f"Output shape = {op.shape}")
-        print("Stats on output data for random normal input:")
-        print(tb.pd.DataFrame(op.numpy().flatten()).describe())
-        print("----------------------------------------", '\n\n')
+        if verbose:
+            print("============  Build Test ==============")
+            print(f"Input shape = {ip.shape}")
+            print(f"Output shape = {op.shape}")
+            print("Stats on output data for random normal input:")
+            print(tb.pd.DataFrame(np.array(op).flatten()).describe())
+            print("----------------------------------------", '\n\n')
 
 
 class Ensemble(tb.Base):
@@ -589,9 +628,9 @@ class Ensemble(tb.Base):
         if hp_class and data_class and model_class:
             self.models = tb.List()
             # only generate the dataset once and attach it to the ensemble to be reused by models.
-            self.data = self.data_class(hp)
+            self.data = self.data_class(hp_class())
             print("Creating Models".center(100, "="))
-            for i in tqdm(range(n)):
+            for i in tqdm(range(size)):
                 hp = self.hp_class()
                 hp.exp_name = str(hp.exp_name) + f'__model__{i}'
                 datacopy = tb.copy.copy(self.data)  # shallow copy
@@ -620,7 +659,7 @@ class Ensemble(tb.Base):
             self.models[i].fit(**kwargs)
             self.performance.append(self.models[i].evaluate(idx=slice(0, -1), viz=False))
             if save:
-                amodel.save_class()
+                self.models[i].save_class()
                 self.performance.save_pickle(self.hp_class.save_dir / "performance.pkl")
         print("\n\n", f" Finished fitting the ensemble ".center(100, ">"), "\n")
 
@@ -650,6 +689,7 @@ def get_mean_max_error(tf):
 
         def reset_states(self):
             self.mme.assign(0.0)
+
     return MeanMaxError
 
 
